% File man/mdp.Rd
\name{mdp}
\alias{mdp}
%- Also NEED an '\alias' for EACH other topic documented here.
\title{MDP -- Dirichlet process mixture of normals}
\description{
  Fits a semiparametric Dirichlet process mixture of normals.
}
\usage{
mdp(n = NULL, p = NULL,
    n.iter = 10000, n.discard = 1000, n.reinit = 10000,
    n.batch = 100, n.predupdate = 100, n.printallpars = 1000,
    m.prior = F, B.prior = F, verbose = 3,
    pxy = F, exy = F, ex = F,
    s = 15, S.init = NULL, q = 5, R = NULL, B.init = NULL,
    cc = 5, C = NULL, m.init = NULL, a = NULL, A = NULL,
    alpha = 1, a0 = 1, b0 = 1,
    k0 = NULL, Y = NULL)
}
%- maybe also 'usage' for other objects documented here.
\arguments{
  \item{n}{number of data records  }
  \item{p}{dimension of each data record }
  \item{n.iter}{number MCMC iterations }
  \item{n.discard}{initial transient}
  \item{n.reinit}{reinitialize every \code{n.reinit} iterations
    (not normally used) }
  \item{n.batch}{save imputed parameters every \code{n.batch} iterations}
  \item{n.predupdate}{save posterior predictive summaries every
    \code{n.predupdate} iterations }
  \item{n.printallpars}{print all parameters every
    \code{n.printallpars} iterations }
  \item{m.prior}{indicator for resampling (1) versus
    fixing (0) the mean of the base measure, m}
  \item{B.prior}{indicator for resampling (1) versus
    fixing (0) the variance-covariance matrix of the base measure, B}
  \item{verbose}{level of comments}
  \item{pxy}{indicator for evaluating posterior predictive
    \eqn{p(x,y \mid data)} for a future observation. Indices
    of \eqn{x,y} are given in \code{ix} and \code{iy}}
  \item{exy}{indicator for evaluating posterior predictive
    expectation \eqn{E(z \mid x,y, data)} for a future observation. Indices
    of \eqn{x,y,z} are given in \code{ix,iy} and \code{iz}}
  \item{ex}{indicator for evaluating posterior predictive
    \eqn{p(x,y \mid data)} for a future observation. Indices
    of \eqn{x,y} are given in \code{ix} and \code{iy}}
  \item{S.init}{initial value for the kernel covariance matrix \eqn{S}}
  \item{s}{degrees of freedom for the inverse Wishart prior on \eqn{S}}
  \item{q}{degrees of freedom for the inverse Wishart
    base measure for \eqn{G_0(V_i)}}
  \item{R}{expectation of the inverse Wishart
    base measure \eqn{G_0(V_i)}}
  \item{R}{ ~~Describe \code{R} here~~ }
  \item{B.init}{initial value for the
    covariance matrix \eqn{B} of the base measure \eqn{G_0(\mu)=N(m,B)}}
  \item{cc}{degrees of freedom of the inverse Wishart
    hyperprior for \eqn{B}}
  \item{C}{mean of the inverse Wishart hyperprior for \eqn{B}}
  \item{m.init}{initial value for the
    mean \eqn{m} of the base measure \eqn{G_0(\mu)=N(m,B)}}
  \item{a}{hyperprior mean for \eqn{m}}
  \item{A}{hyperprior covariance matrix for \eqn{m}}
  \item{alpha}{initial value for the
    total mass parameter \eqn{alpha}}
  \item{a0}{hyperprior parameters for prior on total mass paramter \eqn{\alpha}}
  \item{b0}{hyperprior parameters for prior on total mass paramter \eqn{\alpha}}
  \item{k0}{initial number of distinct clusters}
  \item{Y}{\code{n} by \code{p} data matrix}
}
\details{
  See   \code{\link{mdp-package}}
  for a statement of the probability model.
  The function \code{mdp} initializes and carries out MCMC posterior
  simulation.
  Simulation output is saved in the working directory.
  Change it by using \code{setwd} if desired.
}
\value{
  The function returns no value.
  Simulation output is written to files. 
}
\references{
  the package uses the parametrization defined in:
  
    MacEachern, S.N. and Mueller, P. (1998).
  ``Estimating Mixture of Dirichlet Process Models,''
 Journal of Computational and Graphical Statistics, 7, 223--239.
}
\author{Peter Mueller \email{pm@math.utexas.edu}}
\note{Careful, \code{mdp} writes temporary files
  into the current working directory. The same files
  are used by \code{plt.pxy}, \code{plt.ex} and \code{plt.exy}
  to plot posterior predictive distributions and expectations.
}
\seealso{See also the \code{\link{DPpackage}}, at

  \url{http://student.kuleuven.be/~s0166452/software.html}.}
\examples{
\dontrun{
## Data from Lubischew, A. (1962), "On the use of discriminant functions
##           in taxonomy," Biometrics, 18, 455-477.
data.dir <- system.file("demo",package="mdp")
beetles <- file.path(data.dir,"beetle.data")
Y <- read.table(beetles)   # beetle data. The last column reports species
p <- ncol(Y)         
Y <- as.matrix(Y[,-p]) # don't use the species indicator

## run MCMC
##
mdp(Y=Y,
    pxy=1,exy=1,ex=1,ix=1,iy=2,iz=3,
    q=10,cc=10)

## plot results
##
plt.pxy(img=T,lg=T)         ## biv density estimate p(x,y), 
                            ## x=1st col (ix), y=2nd column (iy)
points(Y[,1],Y[,2],pch=19)  # add data points

plt.pxy(img=F,lg=F)         # same as a contour plot
points(Y[,1],Y[,2],pch=19)

plt.ex(sd=T,zlab="f(x)=E(z | x)")   ## E(z | x, data), for z=3rd col (iz)
plt.ex(sim=T,zlab="f~p(f | data)")  ## show random draws from p(f | data)

plt.exy()                   ## E(z | x,y, data)

## note: for marginals on different coordinates need to
## re-run mdp with new values for ix,iy,iz
}
}

