/* ===================================================================== 
   Predictives and cond expectations in MDP

   to be used with mdp.c in the same directory
/* ===================================================================== */ 
#include<stdio.h>
#include<stdlib.h>
#include<math.h>

#include "rand.h"
#include "nrutil.h"
#include "matrix.h"
#include "vector.h"
#include "rand-nr.h" 
#include "mess.h"
#include "interface.h"
#include "mdp2.h"
#include "predict.h"

/* ***********************************************************
   data structures
/* *********************************************************** */

/* -----------------------------------------------------------
   grid for univariate pdf and exp value
/* ----------------------------------------------------------- */
struct X_GRID{
  int nx;
  int nz;
  int xi,xzi[2],zi;
                       /* index of x var and z var */
  int n_update;        /* number of times z-value has been upd */
  double *x;           /* grid on x - if applicable */
  double *z;           /* grid on x - z values */
  double *ez;           /* grid on x - Ez values */
  double *p;           /* grid on p values if applic */  
  
  int xlist;           /* indicator for evaluating 
			  pdf on mdp data */
  double *pX;          /* z values on mdp data */
};

struct X_GRID *xalloc(void)
/* allocates memory for an X_GRID */
{
  return (struct X_GRID *) malloc(sizeof(struct X_GRID));
}

/* -----------------------------------------------------------
   grid for bivariate  exp value
/* ----------------------------------------------------------- */
struct XY_GRID{
  int nx, ny;          /* size of x and y grids */
  int n_update;        /* number of times z-value has been upd */
  int xi,yi,*xyzi,zi;
                       /* index of x var, y var and z var */
  double *x;           /* grid on x - if applicable */
  double *y;           /* grid on y */
  double **z;          /* matrix over x and y grids */
};

struct XY_GRID *xyalloc(void)
/* allocates memory for an XY_GRID */
{
  return (struct XY_GRID *) malloc(sizeof(struct XY_GRID));
}

/* -----------------------------------------------------------
   grid for bivariate pred 
/* ----------------------------------------------------------- */
struct PXY_GRID{
  int nx, ny;          /* size of x and y grids */
  int n_update;        /* number of times z-value has been upd */
  int xi,yi,*xyzi,zi;
                       /* index of x var, y var and z var */
  int mu;              /* indicator for predictive on th[n+1]
			  (rather than y[n+1] */
  double *x;           /* grid on x - if applicable */
  double *y;           /* grid on y */
  double **z;          /* matrix over x and y grids */
  double **newz;

  int xlist;            /* indicator whether X-list */
  double *pX;           /* predictive on X list */

  double hpd_alpha;
  double *mean;
  double *mode1;
  double *mode2;
  double *up;
  double *down;
};

struct PXY_GRID *pxyalloc(void)
/* allocates memory for an PXY_GRID */
{
  return (struct PXY_GRID *) malloc(sizeof(struct PXY_GRID));
}


/* -----------------------------------------------------------
   grid for mv pred and exp value
/* ----------------------------------------------------------- */
struct XVECY_GRID{
  int nx, ny;          /* size of x and y grids */
  int n_update;        /* number of times z-value has been upd */
  int px;              /* dim of x vec */
  int *xi,yi,zi;       /* index of x var, y var and z var */
  int *xyzi;
  double **x;
  double *y;           /* grid on y */
  double **z;          /* matrix over x and y grids */
};

struct XVECY_GRID *xvecyalloc(void)
/* allocates memory for an XVECY_GRID */
{
  return (struct XVECY_GRID *) malloc(sizeof(struct XVECY_GRID));
}

/* ***********************************************************
   predictive pdf p(x,y)
/* *********************************************************** */

/* -----------------------------------------------------------
   initialize prediction and par estimation
/* ----------------------------------------------------------- */
struct PXY_GRID *pxy_init(struct MDP_PARS mdp)
{
  struct PXY_GRID *pxy;
  int nx, ny, xi, yi, mu, j;
  double hpd_alpha,lox,loy,hix,hiy;

  openIn("pxy-init.mdp");
  scanInt("nx", &nx);
  scanInt("ny", &ny);
  scanInt("xi", &xi);
  scanInt("yi", &yi);
  scanDouble("hpd-alpha", &hpd_alpha);
  scanInt("mu", &mu);

  /* allocate memory */
  pxy = pxyalloc();
  pxy->x = dvector(0,nx-1);
  pxy->y = dvector(0,ny-1);
  pxy->z = dmatrix(0,nx-1,0,ny-1);
  pxy->newz = dmatrix(0,nx-1,0,ny-1);
  pxy->xyzi = ivector(0,1);
  pxy->mean = dvector(0,nx-1);
  pxy->mode1 = dvector(0,nx-1);
  pxy->mode2 = dvector(0,nx-1);
  pxy->up = dvector(0,ny-1);
  pxy->down = dvector(0,ny-1);

  /* initialize members */
  pxy->nx = nx;
  pxy->xi = xi;
  pxy->yi = yi;
  pxy->xyzi[0] = xi;
  pxy->xyzi[1] = yi;
  pxy->ny = ny;
  pxy->n_update = 0;
  pxy->hpd_alpha = hpd_alpha;
  pxy->mu = mu;

  /* read in x and y grid */
  scanDouble("lox",&lox);
  scanDouble("hix",&hix);
  scanDouble("loy",&loy);
  scanDouble("hiy",&hiy);
  for (j=0;j<nx;j++)
    pxy->x[j] = lox + j*(hix-lox)/(1.0*nx);
  for (j=0;j<ny;j++)
    pxy->y[j] = loy + j*(hiy-loy)/(1.0*ny);

  /* X list */
  scanInt("xlist", &pxy->xlist);
  if (pxy->xlist == 1)
    pxy->pX = dvector(0,mdp.n_obs);
  closeIn();
  return pxy;
}

/* -----------------------------------------------------------
   print_pxy
/* ----------------------------------------------------------- */
void print_pxy(struct PXY_GRID *pxy, struct MDP_PARS mdp)
{
  int i;
  FILE *hpdFile;

  /* print x-grid */
  if (pxy->mu==0) openOut("pxy-x.mdp");
  else openOut("pxy-mux.mdp");
  writeDoubleArray(pxy->x,pxy->nx,1);
  closeOut();

  /* write y-grid */
  if (pxy->mu==0) openOut("pxy-y.mdp");
  else openOut("pxy-muy.mdp");
  writeDoubleArray(pxy->y,pxy->ny,1);
  closeOut();

  /* write z-grid */
  if (pxy->mu==0) openOut("pxy-z.mdp");
  else openOut("pxy-muz.mdp");
  writeDoubleMatrix2(pxy->z,pxy->nx,pxy->ny);
  closeOut();

  /* write newz-grid */
  if (pxy->mu==0) openOut("pxy-newz.mdp");
  else openOut("pxy-newmuz.mdp");
  writeDoubleMatrix2(pxy->newz,pxy->nx,pxy->ny);
  closeOut();

  /* write out PX.mdp */
  if (pxy->xlist == 1){
    openOut("pxy-pX.mdp");
    writeDoubleArray(pxy->pX,1,mdp.n_obs);
    closeOut();
  }
}
/* -----------------------------------------------------------
   hpd-regions
/* ----------------------------------------------------------- */
void pxy_hpd(struct PXY_GRID *pxy, struct MDP_PARS mdp)
{
  int i,j,jdown,jup,jmax, p,nx,ny;
  double max, *z, sum, total, alpha;
  
  /* set aux pars */
  p = mdp.p;
  nx = pxy->nx;
  ny = pxy->ny;
  alpha = pxy->hpd_alpha;

  for (i=0;i<nx;i++){
    z = pxy->z[i];
    /* find 1st local max */
    for (sum=0 ,j=1; (z[j-1] < z[j]) & j<ny ; j++);
    pxy->mode1[i] =  pxy->y[j-1];
    /* goto valey */
    for (; (z[j-1] > z[j]) & j<ny ; j++);
    /* find 2nd local mode */
    for (; (z[j-1] < z[j]) & j<ny ; j++);
    pxy->mode2[i] = pxy->y[j-1];
    /* find global mode & sum */
    for(max = sum = pxy->mean[i]=0.0,j=0; j<ny; j++){
      pxy->mean[i] += pxy->y[j]*z[j];
      sum += z[j];
      if (z[j] > max){
	max = z[j];
	jmax = j;
      }
    }
    pxy->mean[i] /= sum;
    total = z[jmax];
    for (jup=jmax, jdown=jmax; total < alpha*sum; ){
      if (jup > ny-2){
	jdown -= 1;
	total += z[jdown];
      }
      else if ( (z[jup+1] > z[jdown-1]) | (jdown < 2) ){
	jup += 1;
	total += z[jup];
      }
      else{
	jdown -= 1;
	total += z[jdown];
      }
    }
    pxy->up[i] = pxy->y[jup];
    pxy->down[i] = pxy->y[jdown];
  }
}
      

/* -----------------------------------------------------------
   pxy_updatae
/* ----------------------------------------------------------- */
/*   get predicted values over grid xgrid, ygrid (global??) */
void pxy_update(int time, struct PXY_GRID *pxy,
	     struct MDP_PARS mdp)
{
  int i, j, k, p, xi, yi, nx, ny, *xyzi, n_updates;
  double y[2], r, **z, *pX, **newz, kd;
  double **V, **V_inv, **L, m[2];
  double sum_wgt, sum_newwgt, alpha, an;

  /* set aux pars */
  p = mdp.p;
  nx = pxy->nx;
  ny = pxy->ny;
  xyzi = pxy->xyzi;
  xi = pxy->xi;
  yi = pxy->yi;
  alpha = mdp.alpha;
  sum_wgt = 0.0;
  kd = 1.0*mdp.n_class;
  an = mdp.alpha + mdp.n_obs*1.0;

  /* allocate memory */
  z = dmatrix(0,nx,0,ny);
  pX = dvector(0,mdp.n_obs);
  newz = dmatrix(0,nx,0,ny);
  V = dmatrix(0,1,0,1);
  V_inv = dmatrix(0,1,0,1);
  L = dmatrix(0,1,0,1);

  /* -----------------------------------------------------------
     predictive if y is from G0
  /* ----------------------------------------------------------- */
  /* initialize L=V^(-1/2) for Snew = (1+tau)*s/(s-1)*S */

  if (pxy->mu == 1)
    dsubmatrix(mdp.Vj[mdp.n_class],p,p,V,2,2,xyzi,xyzi);
  else  dsubmatrix(mdp.Vt[mdp.n_class],p,p,V,2,2,xyzi,xyzi);
  d_invert(2,V,V_inv);
  d_chol_decomp(2,V_inv,L);
  dsubvector(mdp.mean,p,m,2,xyzi);
        
  /* loop over all grid points - compute predictive conditional
     on current pars */
  for (i = 0; i < nx; i++){
    y[0] = pxy->x[i];
    for (j = 0; j < ny; j++){
      y[1] = pxy->y[j];
      z[i][j] = alpha/an*mvn_pdf(y, m, L, 2);
      newz[i][j] = 0.0;
    }
  }
  if (pxy->xlist == 1){
    for (i=0; i < mdp.n_obs; i++){
      y[0] = mdp.Y[i][xi]; 
      y[1] = mdp.Y[i][yi]; 
      pX[i] = alpha/an * mvn_pdf(y,m,L,2);
    }
  }

  /* -----------------------------------------------------------
     pred if y is one of the existing classes 
  /* ----------------------------------------------------------- */
  for (k = 0; k < mdp.n_class; k++){
    dsubvector(mdp.mj[k],p,m,2,xyzi);
    if (pxy->mu == 1)
      dsubmatrix(mdp.Vj[k],p,p,V,2,2,xyzi,xyzi);
    else dsubmatrix(mdp.Vt[k],p,p,V,2,2,xyzi,xyzi);
    d_invert(2,V,V_inv);
    d_chol_decomp(2,V_inv,L);
    r = mdp.count[k]*1.0;
    
    /* loop over all grid points - compute predictive conditional
       on current pars */
    for (i = 0; i < nx; i++){
      y[0] = pxy->x[i];
      for (j = 0; j < ny; j++){
	y[1] = pxy->y[j];
	z[i][j] += r/an*mvn_pdf(y,m,L,2);
	newz[i][j] += r/an*mvn_pdf(y,m,L,2);
      }
    }
    if (pxy->xlist == 1){
      for (i=0; i < mdp.n_obs; i++){
	y[0] = mdp.Y[i][xi]; 
	y[1] = mdp.Y[i][yi];
	pX[i] += r/an* mvn_pdf(y,m,L,2);
    }
  }

  }/* for k */

  /* add new z to pxy->z */
  for(sum_wgt=sum_newwgt=0.0,i=0;i<nx;i++)
    for(j=0;j<ny;j++){
      sum_newwgt += newz[i][j];
      sum_wgt += z[i][j];
    }
  n_updates = pxy->n_update;
  for (i = 0; i < nx; i++)
    for (j = 0; j < ny; j++){
      pxy->z[i][j] = n_updates*pxy->z[i][j]/(n_updates+1.0) + 
	z[i][j]/(sum_wgt*(n_updates+1.0));
      pxy->newz[i][j] = newz[i][j]/sum_newwgt;
    }
  /* add new pX to pxy->pX */
  if (pxy->xlist == 1)
    for(i=0; i<mdp.n_obs; i++)
      pxy->pX[i] = (n_updates* pxy->pX[i] + pX[i])/(n_updates+1.0);
  /* increment n_update */
  pxy->n_update = n_updates+1;

  /* free memory */
  free_dmatrix(z,0,nx,0,ny);
  free_dmatrix(newz,0,nx,0,ny);
  free_dmatrix(V,0,1,0,1);
  free_dmatrix(V_inv,0,1,0,1);
  free_dmatrix(L,0,1,0,1);
  free_dvector(pX,0,mdp.n_obs);

  return;
}        
                

/* ***********************************************************
   bivariate conditional expectation
/* *********************************************************** */

/* -----------------------------------------------------------
   initialize biv expected values
/* ----------------------------------------------------------- */
struct XY_GRID *exy_init()
{
  int nx,ny,xi,yi,zi;
  struct XY_GRID *exy;

  /* read in pars */
  openIn("exy-init.mdp");
  scanInt("nx", &nx);
  scanInt("ny", &ny);
  scanInt("xi", &xi);
  scanInt("yi", &yi);
  scanInt("zi", &zi);

  /* allocate mem */
  exy = xyalloc();
  exy->x = dvector(0,nx-1);
  exy->y = dvector(0,ny-1);
  exy->z = dmatrix(0,nx-1,0,ny-1);
  exy->xyzi = ivector(0,2);
  exy->n_update = 0;

  /* initialize members */
  exy->nx = nx;
  exy->ny = ny;
  exy->xi = exy->xyzi[0] = xi;
  exy->yi = exy->xyzi[1] = yi;
  exy->zi = exy->xyzi[2] = zi;

  /* read in grids for x and y */
  scanDoubleArray("xgrid", exy->x, nx);
  scanDoubleArray("ygrid", exy->y, ny);

  closeIn();
  return exy;
}

/* -----------------------------------------------------------
   print_exy
/* ----------------------------------------------------------- */
/*   print E(z|xy) */
void print_exy(struct XY_GRID *exy)
{
  /* print x-grid */
  openOut("exy-x.mdp");
  writeDoubleArray(exy->x,exy->nx,1);
  closeOut();
  /* write y-grid */
  openOut("exy-y.mdp");
  writeDoubleArray(exy->y,exy->ny,1);
  closeOut();

  /* write z-grid */
  openOut("exy-z.mdp");
  writeDoubleMatrix2(exy->z,exy->nx,exy->ny);
  closeOut();
}

/* -----------------------------------------------------------
   update
/* ----------------------------------------------------------- */
/*   get expected values over grid xgrid, ygrid (global??) */
void exy_update(int time, struct XY_GRID *exy, 
		struct MDP_PARS mdp)
{
  int i, j, k, p, nx, ny, zi, *xyzi, n_updates;
  double y[2], e[2], Le[2], dz, r, alpha, coeff, Ek, **sum_wgt;
  double **z, m[3], **L, **V, **V_inv, Vxy_z[2], an, kd;

  /* set aux pars */
  p = mdp.p;
  nx = exy->nx;
  ny = exy->ny;
  zi = exy->zi;
  xyzi = exy->xyzi;
  alpha = mdp.alpha;
  kd = 1.0*mdp.n_class;
  an = mdp.alpha + mdp.n_obs*1.0;

  /* allocate mem */
  z = dmatrix(0,nx,0,ny);
  sum_wgt = dmatrix(0,nx,0,ny);
  V = dmatrix(0,2,0,2);
  V_inv = dmatrix(0,1,0,1);
  L = dmatrix(0,1,0,1);

  /* -----------------------------------------------------------
     predictive if y from G0 */
  /* ----------------------------------------------------------- */
  /* initialize L for V = (1+tau)*s/(s-1)*S */
  dsubmatrix(mdp.Vt[mdp.n_class], p,p,V, 3,3,xyzi,xyzi);
  d_invert(2,V,V_inv);
  d_chol_decomp(2,V_inv,L);
  dsubvector(mdp.Vt[mdp.n_class][zi],p,Vxy_z,2,xyzi);

  /* loop over all pixels */
  dsubvector(mdp.mean,p,m,3,xyzi);
  for (i = 0; i < nx; i++){
    y[0] = exy->x[i];
    for (j = 0; j < ny; j++){
      y[1] = exy->y[j];
      coeff = mvn_pdf(y, m, L, 2);
      x_min_y(y,m,e,2);
      Ax(V_inv,e,Le,2);
      xy(Le,Vxy_z,&dz,2);
      Ek = m[2]+dz;
      z[i][j] = alpha/an*coeff*Ek;
      sum_wgt[i][j] = alpha/an*coeff;
      if (sum_wgt[i][j] == 0){
	/* this is a spy */
	printf("\n WARNING: 0 coeff \n");
      }
    }
  }

  /* pred if y is in one of the existing classes */
  for (k = 0; k < mdp.n_class; k++){
    r = 1.0*mdp.count[k];
    dsubvector(mdp.mj[k],p,m,3,xyzi);
    dsubmatrix(mdp.Vt[k],p,p,V,3,3,xyzi,xyzi);
    d_invert(2,V,V_inv);
    d_chol_decomp(2,V_inv,L);
    dsubvector(mdp.Vt[k][zi],p,Vxy_z,2,xyzi);
     for (i = 0; i < nx; i++){
      y[0] = exy->x[i];
      for (j = 0; j < ny; j++){
	y[1] = exy->y[j];
	coeff = mvn_pdf(y, m, L, 2);
	/* Ek = mu_z + V[x,z]*(ynew-mu[x])/V[z,z] */
	x_min_y(y,m,e,2);
	Ax(V_inv,e,Le,2);
	xy(Le,Vxy_z,&dz,2);
	Ek = m[2]+dz;
	z[i][j] += r/an*coeff*Ek;
	sum_wgt[i][j] += r/an*coeff;
	}
    }
  }

  /* loop over all grid points - compute predictive conditional
     on current pars */
  n_updates = exy->n_update;
  for (i = 0; i < nx; i++){
    for (j = 0; j < ny; j++){
      if (sum_wgt[i][j] <= 0.0){
	printf("At i= %d, j=%d \n",i,j);
	error3("pxy_update: non-posistive weights.");
      }
      z[i][j] /= sum_wgt[i][j];
      exy->z[i][j] = n_updates*exy->z[i][j]/(n_updates+1.0) + 
	z[i][j]/(n_updates+1.0);
    }
  }
  exy->n_update = n_updates+1;

  /* free memory */
  free_dmatrix(sum_wgt,0,nx,0,ny);
  free_dmatrix(z,0,nx,0,ny);
  free_dmatrix(V,0,2,0,2);
  free_dmatrix(V_inv,0,1,0,1);
  free_dmatrix(L,0,1,0,1);

  return;
}        

/* ***********************************************************
   mv conditional expectation
/* *********************************************************** */

/* -----------------------------------------------------------
   initialize mv expected values
/* ----------------------------------------------------------- */
struct XVECY_GRID *exvecy_init(int nplt, struct MDP_PARS mdp)
{
  int ny, nx, px, *xi, yi, zi, p, i, n_updates;

  struct XVECY_GRID *exy;

  /* read pars */
  openIn("exvecy-init.mdp");
  scanInt("px", &px);
  xi = ivector(0,px-1);
  scanIntArray("xi",xi, px);
  scanInt("yi",&yi);
  scanInt("zi",&zi);
  if (nplt>1){
    scanInt("nx",&nx);
  }
  else nx = mdp.n_obs;
  scanInt("ny", &ny);

  /* aux pars */
  p = mdp.p;

  /* allocate mem */
  exy = xvecyalloc();
  exy->xyzi = ivector(0,px+1);
  exy->x = dmatrix(0,nx-1,0,px-1);
  exy->y = dvector(0,ny-1);
  exy->z = dmatrix(0,nx-1,0,ny-1);

  /* initialize ygrid */
  exy->ny = ny;
  exy->yi = yi;
  scanDoubleArray("ygrid", exy->y, ny);

  /* initialize xvec grid */
  exy->px = px;
  exy->nx = nx;
  if (nplt==1){
    for(i=0;i<nx;i++){
      dsubvector(mdp.Y[i],p,exy->x[i],px,xi);
    }
  }
  else{
    scanDoubleMatrix("xgrid",exy->x, nx, px);
  }

  /* initialize z matrix */
  exy->zi = zi;
  for(i=0;i<px;i++) exy->xyzi[i] = xi[i];
  exy->xyzi[px] = yi;
  exy->xyzi[px+1] = zi;

  exy->n_update = 0;

  closeIn();
  free_ivector(xi,0,px-1);
  return exy;
}

/* -----------------------------------------------------------
   print_exvecy
/* ----------------------------------------------------------- */
/*   print E(z|xvec,y) */
void print_exvecy(struct XVECY_GRID *exy, int pltn)
{
  
  /* write y-grid */
  if(pltn==1) openOut("exvecy-y.mdp");
  else openOut("exvecy2-y.mdp");
  writeDoubleArray(exy->y,exy->ny,1);
  closeOut();

  /* write x-grid */
  if(pltn==1) openOut("exvecy-x.mdp");
  else openOut("exvecy2-x.mdp");
  writeDoubleMatrix2(exy->x,exy->nx,exy->px);
  closeOut();

  /* write z-grid */
  if(pltn==1) openOut("exvecy-z.mdp");
  else openOut("exvecy2-z.mdp");
  writeDoubleMatrix2(exy->z,exy->nx,exy->ny);
  closeOut();
}


/* -----------------------------------------------------------
   update
/* ----------------------------------------------------------- */
/*   get expected values over grid xgrid, ygrid (global??) */
void exvecy_update(int time, struct XVECY_GRID *exy, 
		struct MDP_PARS mdp)
{
  int i, j, k, h, nx, ny, p, px, pxy, pxyz, *xyzi, zi, n_updates;
  double exy_c, *y, *m,**V, **V_inv, **L, **z, *Vxy_z,
         *e, *Le, r, alpha, coeff, Ek, **sum_wgt, dz, kd, an;

  /* set aux pars */
  p = mdp.p;
  nx = exy->nx;
  ny = exy->ny;
  px = exy->px;
  pxy = px+1;
  pxyz = px+1+1;
  xyzi = exy->xyzi;
  zi = exy->zi;
  alpha = mdp.alpha;
  kd = mdp.n_class*1.0;
  an = mdp.alpha + mdp.n_obs*1.0;

  /* alloc memory */
  m = dvector(0,pxyz-1);
  L = dmatrix(0,pxy-1,0,pxy-1);
  V_inv = dmatrix(0,pxy-1,0,pxy-1);
  V = dmatrix(0,pxyz-1,0,pxyz-1);
  Vxy_z = dvector(0,pxy-1);
  e = dvector(0,pxy-1);
  Le = dvector(0,pxy-1);
  y = dvector(0,pxy-1);
  z = dmatrix(0,nx-1,0,ny-1);
  sum_wgt = dmatrix(0,nx-1,0,ny-1);

  /* predictive if y from G0 */
  /* initialize L for V = (1+tau)*s/(s-1)*S and m = mean */
  dsubmatrix(mdp.Vt[mdp.n_class],p,p,V,pxyz,pxyz,xyzi,xyzi);
  d_invert(pxy,V,V_inv);
  d_chol_decomp(pxy,V_inv,L);
  dsubvector(mdp.mean,p,m,pxyz,xyzi);
  dsubvector(mdp.Vt[mdp.n_class][zi],p,Vxy_z,pxy,xyzi);

  /* loop over all x/y pixels */
  for (i = 0; i < nx; i++){
    y_assgn_x(y,exy->x[i],px);
    /* note: exy->x is already subset and ordererd */
    /* dsubvector(exy->x[i],p,y,px,xyzi); OLD & wrong?? */
    for (j = 0; j < ny; j++){
      y[pxy-1] = exy->y[j];
      coeff = mvn_pdf(y, m, L, pxy);
      x_min_y(y,m,e,pxy);
      Ax(V_inv,e,Le,pxy);
      xy(Le,Vxy_z,&dz,pxy);
      Ek = m[pxyz-1] + dz;
      z[i][j] = alpha/an*coeff*Ek;
      sum_wgt[i][j] = alpha/an*coeff; 
    }
  }

  /* pred if y is in one of the existing classes */
  for (k = 0; k < mdp.n_class; k++){
    /* set up m, V, V_inv and L = (chol-decomp V)^-1 */
    r = 1.0*mdp.count[k];
    dsubvector(mdp.mj[k],p,m,pxyz,xyzi);
    dsubmatrix(mdp.Vt[k],p,p,V,pxyz,pxyz,xyzi,xyzi);
    d_invert(pxy,V,V_inv);
    d_chol_decomp(pxy,V_inv,L);
    dsubvector(mdp.Vt[k][zi],p,Vxy_z,pxy,xyzi);

    /* loop over all pixels */
    for (i = 0; i < nx; i++){
      y_assgn_x(y,exy->x[i],px);
      for (j = 0; j < ny; j++){
	y[pxy-1] = exy->y[j];
	coeff = mvn_pdf(y, m, L, pxy);
	/* Ek = mu_z + V[x,z]*(ynew-mu[x])/V[z,z] */
	x_min_y(y,m,e,pxy);
	Ax(V_inv,e,Le,pxy);
	xy(Le,Vxy_z,&dz,pxy);
	Ek = m[pxyz-1] + dz;
	z[i][j] += r/an*coeff*Ek;
	sum_wgt[i][j] += r/an*coeff;
	}
    }
  }

  /* loop over all grid points - compute predictive conditional
     on current pars */
  n_updates = exy->n_update;
  for (i = 0; i < nx; i++)
    for (j = 0; j < ny; j++){
      if (sum_wgt[i][j] <= 0.0){
	printf("\ni=%d, j=%d ",i,j);
	error("print exvecy", "exvecy_update: non-positive weight.",1);
      }
      z[i][j] /= sum_wgt[i][j];
      exy->z[i][j] = n_updates*exy->z[i][j]/(n_updates+1.0) + 
	z[i][j]/(n_updates+1.0);
    }
  exy->n_update = n_updates+1;

  /* free mem */
  free_dvector(Vxy_z,0,pxy-1);
  free_dvector(m,0,pxyz-1);
  free_dmatrix(L,0,pxy-1,0,pxy-1);
  free_dmatrix(V_inv,0,pxy-1,0,pxy-1);
  free_dmatrix(V,0,pxyz-1,0,pxyz-1);
  free_dvector(y,0,pxy-1);
  free_dmatrix(z,0,nx-1,0,ny-1);
  free_dmatrix(sum_wgt,0,nx-1,0,ny-1);
  free_dvector(e,0,pxy-1);
  free_dvector(Le,0,pxy-1);

  return;
}        
                

/* ***********************************************************
   univariate cond exp
/* *********************************************************** */
/* -----------------------------------------------------------
   initialize univ expected values
/* ----------------------------------------------------------- */
struct X_GRID *ex_init()
{
  int nx, nz, xi, zi;
  struct X_GRID *ex;

  /* read pars */
  openIn("ex-init.mdp");
  scanInt("nx", &nx);
  scanInt("nz", &nz);
  scanInt("xi", &xi);
  scanInt("zi", &zi);

  /* alloc mem */
  ex = xalloc();
  ex->x = dvector(0,nx-1);
  ex->z = dvector(0,nx-1);
  ex->ez = dvector(0,nx-1);

  /* init members */
  ex->nx = nx; ex->nz = nz;
  ex->xi = ex->xzi[0] = xi;
  ex->zi = ex->xzi[1] = zi;
  ex->n_update = 0;

  /* read x grid */
  scanDoubleArray("xgrid", ex->x, nx);

  closeIn();
  return ex;
}

/* -----------------------------------------------------------
   print_ex
/* ----------------------------------------------------------- */
/*   print E(z|xy) */
void print_ex(struct X_GRID *ex)
{
  int nupd;

  nupd = ex->n_update;

  /* print x-grid */
  if (nupd == 1) {
    openOut("ex-x.mdp");
    writeDoubleArray(ex->x,ex->nx,1);
    closeOut();
  }

  /* write z-grid */
  if (nupd == 1) openOut("ex-z.mdp");
  else openAppend("ex-z.mdp");
  writeDoubleArray(ex->z,1,ex->nx);
  closeOut();

  openOut("ex-ez.mdp");
  writeDoubleArray(ex->ez,1,ex->nx);
  closeOut();
}

/* -----------------------------------------------------------
   expect
/* ----------------------------------------------------------- */
/*   get expected values over grid xgrid, ygrid (global??) */
     void ex_update(int time, struct X_GRID *ex, 
		struct MDP_PARS mdp)
{
  int i, j, k, p, nx, xi, zi, n_updates;
  double ex_c, y, sxz, sxx, m, mz, sx, r, *z, kd;
  double coeff, Ek, alpha, *sum_wgt,an;

  /* set aux pars */  
  p = mdp.p;
  xi = ex->xi;
  zi = ex-> zi;
  nx = ex->nx;
  alpha = mdp.alpha;
  kd = 1.0*mdp.n_class;
  an = alpha + mdp.n_obs*1.0;

  /* alloc mem */
  z = dvector(0,nx-1);
  sum_wgt = dvector(0,nx-1);
  
  /* initialize L for V = (1+tau)*s/(s-1)*S */
  sxz = mdp.Vt[mdp.n_class][xi][zi];
  sxx = mdp.Vt[mdp.n_class][xi][xi];
  sx = sqrt(sxx);
  m = mdp.mean[xi];
  mz = mdp.mean[zi];

  /* predictive if y from G0 */
  for (i = 0; i < nx; i++){
    y = ex->x[i];
    coeff = pdfnorm(y, m, sx);
    Ek = mz + sxz*(y-m)/sxx;
    z[i] = alpha/an*coeff*Ek;
    sum_wgt[i] = alpha/an*coeff;
  }
  
  
  /* pred if y is in one of the existing classes */
  for (k = 0; k < mdp.n_class; k++){
    r = 1.0*mdp.count[k];
    m = mdp.mu[k][xi];
    sxz = mdp.Vt[k][xi][zi];
    sxx = mdp.Vt[k][xi][xi];
    mz = mdp.mj[k][zi];
    sx = sqrt(sxx);
    for (i = 0; i < nx; i++){
      y = ex->x[i];
      coeff = pdfnorm(y, m, sx);
      /* Ek = mu_z + V[x,z]*(ynew-mu[x])/V[z,z] */
      Ek = mz + sxz*(y-m)/sxx;
      z[i] += r/an*coeff*Ek;
      sum_wgt[i] += r/an*coeff;
    }
  }
  
      
  n_updates = ex->n_update+1;
  if (n_updates <= ex->nz)
    j = n_updates;
  else j = ex->nz;

  for (i = 0; i < nx; i++){
    z[i] /= sum_wgt[i];
    ex->z[i] = z[i];
    ex->ez[i] = (z[i] + ex->ez[i]*(n_updates-1))/n_updates;
  }
  ex->n_update = n_updates;

  /* free alloc mem */
  free_dvector(z,0,nx-1);
  free_dvector(sum_wgt,0,nx-1);
  return;
}        
                

/* ***********************************************************
   univariate predictive
/* *********************************************************** */
/* -----------------------------------------------------------
   initialize univ predicitve
/* ----------------------------------------------------------- */
struct X_GRID *px_init(struct MDP_PARS mdp)
{
  int nx, xi, zi;
  struct X_GRID *px;

  /* read pars */
  openIn("px-init.mdp");
  scanInt("nx", &nx);
  scanInt("xi", &xi);

  /* alloc mem */
  px = xalloc();
  px->x = dvector(0,nx-1);
  px->p = dvector(0,nx-1);
  px->z = dvector(0,nx-1); /* for currently imputed density */

  /* init members */
  px->nx = nx;
  px->xi = px->xzi[0] = xi;
  px->n_update = 0;

  /* read x grid */
  scanDoubleArray("xgrid", px->x, nx);

  /* check for evaluation on mdp data */
  scanInt("xlist", &px->xlist);
  if (px->xlist==1)
    px->pX = dvector(0,mdp.n_obs);
  
  closeIn();
  return px;
}

/* -----------------------------------------------------------
   print_px
/* ----------------------------------------------------------- */
/*   print E(z|xy) */
void print_px(struct X_GRID *px, struct MDP_PARS mdp)
{
  int nupd;

  nupd = px->n_update;

  /* print x-grid */
  openOut("px-x.mdp");
  writeDoubleArray(px->x,px->nx,1);
  closeOut();

  /* write z-grid */
  openOut("px-p.mdp");
  writeDoubleArray(px->p,px->nx,1);
  closeOut();

  /* write z-grid */
  if (nupd == 1) openOut("px-z.mdp");
  else openAppend("px-z.mdp");
  writeDoubleArray(px->z,1,px->nx);
  closeOut();

  /* write pred on mdp data */
  if (px->xlist==1){
    openOut("px-pX.mdp");
    writeDoubleArray(px->pX,mdp.n_obs,1);
    closeOut();
  }
}

/* -----------------------------------------------------------
   update
/* ----------------------------------------------------------- */
/*   get expected values over grid xgrid, ygrid (global??) */
     void px_update(int time, struct X_GRID *px, 
		struct MDP_PARS mdp)
{
  int i, j, k, p, nx, xi, n_updates;
  double y, sxx, m, sx, r, *z, *pX, kd;
  double pi, alpha, an;

  /* set aux pars */  
  p = mdp.p;
  xi = px->xi;
  nx = px->nx;
  alpha = mdp.alpha;
  kd = 1.0*mdp.n_class;
  an = mdp.alpha+1.0*mdp.n_obs;

  /* alloc mem */
  z = px->z;
  pX = dvector(0,mdp.n_obs);
  
  /* initialize L for V = (1+tau)*s/(s-1)*S */
  sxx = mdp.Vt[mdp.n_class][xi][xi];
  sx = sqrt(sxx);
  m = mdp.mean[xi];

  /* predictive if y from G0 */
  for (i = 0; i < nx; i++){
    y = px->x[i];
    pi = pdfnorm(y, m, sx);
    z[i] = alpha/an*pi;
  }
  if (px->xlist==1)
    for(i=0; i<mdp.n_obs;i++){
      y = mdp.Y[i][xi];
      pi = pdfnorm(y,m,sx);
      pX[i] = alpha/an*pi;
    }
  
  /* pred if y is in one of the pxisting classes */
  for (k = 0; k < mdp.n_class; k++){
    r = 1.0*mdp.count[k];
    m = mdp.mu[k][xi];
    sxx = mdp.V[k][xi][xi];
    sx = sqrt(sxx);
    for (i = 0; i < nx; i++){
      y = px->x[i];
      pi = pdfnorm(y, m, sx);
      z[i] += r/an*pi;
    }
    if (px->xlist==1)
      for(i=0; i<mdp.n_obs;i++){
	y = mdp.Y[i][xi];
	pi = pdfnorm(y,m,sx);
	pX[i] += r/an*pi;
      }
  }

  /* update px->p */
  n_updates = px->n_update;
  for (i = 0; i < nx; i++){
    px->p[i] = n_updates*px->p[i]/(n_updates+1.0) + 
      z[i]/(n_updates+1.0);
  }
  /* update px->pX */
  if (px->xlist == 1)
    for(i=0; i<mdp.n_obs;i++)
      px->pX[i] = (n_updates* (px->pX[i]) + pX[i])/(n_updates+1.0);
  
  /* increment n_update */
  px->n_update = n_updates+1;

  /* free alloc mem */
  free_dvector(pX,0,mdp.n_obs);
  return;
}        



/* ===================================================================== 
   problem specific pred routines
/* ===================================================================== */ 

static struct XY_GRID *exy;
static struct PXY_GRID *pxy;
static struct PXY_GRID *mupxy;
static struct XVECY_GRID *exvecy;
static struct XVECY_GRID *exvecy2;
static struct X_GRID *ex;
static struct X_GRID *px;

/* ***********************************************************
   pred_init
/* *********************************************************** */
void pred_init(struct MDP_PARS mdp)
{
  message("pred_init");
  if (mdp.px == 1) px = px_init(mdp);
  if (mdp.ex == 1) ex = ex_init();
  if (mdp.exy == 1) exy = exy_init();
  if (mdp.exvecy == 1) exvecy = exvecy_init(1,mdp);
  if (mdp.exvecy == 2) exvecy2 = exvecy_init(2,mdp);
  if (mdp.pxy == 1) pxy = pxy_init(mdp);
  if (mdp.mupxy == 1){
    mupxy = pxy_init(mdp);
    mupxy->mu = 1;
  }
}

/* ***********************************************************
   pred_update
/* *********************************************************** */
void pred_update(int time, struct MDP_PARS mdp)
{
  if (mdp.px == 1) px_update(time,px,mdp);
  if (mdp.ex == 1) ex_update(time,ex,mdp);
  if (mdp.exy == 1) exy_update(time,exy,mdp);
  if (mdp.exvecy == 1) exvecy_update(time,exvecy,mdp);
  if (mdp.exvecy == 2) exvecy_update(time,exvecy2,mdp);
  if (mdp.pxy == 1) pxy_update(time,pxy,mdp);
  if (mdp.mupxy == 1) pxy_update(time,mupxy,mdp);
  pred_finish(mdp);
}
/* ***********************************************************
   pred_finish
/* *********************************************************** */
void pred_finish(struct MDP_PARS mdp)
{
  if (mdp.px == 1) print_px(px,mdp);
  if (mdp.ex == 1) print_ex(ex);
  if (mdp.exy == 1) print_exy(exy);
  if (mdp.exvecy == 1) print_exvecy(exvecy,1);
  if (mdp.exvecy == 2) print_exvecy(exvecy2,2);
  if (mdp.pxy == 1) print_pxy(pxy,mdp);
  if (mdp.mupxy == 1) print_pxy(mupxy,mdp);
}

/* ***********************************************************
   hpd
/* *********************************************************** */
void make_hpd(struct MDP_PARS mdp)
{
  int i;
  FILE *hpdFile;

  /* make hpd region */
  pxy_hpd(pxy,mdp);

  /* print hpd_region */
  hpdFile = openOut("pxy-hpd.mdp");
  for(i=0; i< pxy->nx; i++){
    fprintf(hpdFile,"%4.2f \t %4.2f \t %4.2f \t %4.2f %4.2f %4.2f\n",
	    pxy->x[i], pxy->down[i], pxy->mean[i], pxy->up[i],
	    pxy->mode1[i], pxy->mode2[i]);
  }
  fclose(hpdFile);
}

